@@ -5,7 +5,7 @@ int be_getsockname(int, struct sockaddr *, int *);
 int be_listen(int, int);
 int be_accept(int, struct sockaddr *, int *);
 int be_bind(int, struct sockaddr *, int);
-int be_pipe(int *);
+int be_pipe(int [2]);
 int be_read(int, void *, int);
 int be_write(int, void *, int);
 int be_close(int);
