@@ -192,6 +192,7 @@ void make_connection(struct connection *c, int port, int *sock, void (*func)(str
 	size_t sl;
 	struct conn_info *b;
 	if (*c->socks_proxy) {
+		/* pasing duplicated in proxy_ok_dialog */
 		unsigned char *p = cast_uchar strchr(cast_const_char c->socks_proxy, '@');
 		if (p) p++;
 		else p = c->socks_proxy;
@@ -199,10 +200,11 @@ void make_connection(struct connection *c, int port, int *sock, void (*func)(str
 		socks_port = 1080;
 		if ((p = cast_uchar strchr(cast_const_char host, ':'))) {
 			long lp;
+			char *end;
 			*p++ = 0;
 			if (!*p) goto badu;
-			lp = strtol(cast_const_char p, (char **)(void *)&p, 10);
-			if (*p || lp <= 0 || lp >= 65536) {
+			lp = strtol(cast_const_char p, &end, 10);
+			if (*end || lp <= 0 || lp >= 65536) {
 				badu:
 				mem_free(host);
 				setcstate(c, S_BAD_URL);
@@ -232,8 +234,9 @@ void make_connection(struct connection *c, int port, int *sock, void (*func)(str
 	log_string(cast_uchar ":");
 	log_number(socks_port != -1 ? socks_port : port);
 	log_string(cast_uchar "\n");
-	if (c->last_lookup_state.addr.n) {
+	if (c->last_lookup_state.addr_index < c->last_lookup_state.addr.n) {
 		b->l.addr = c->last_lookup_state.addr;
+		b->l.addr_index = c->last_lookup_state.addr_index;
 		b->l.dont_try_more_servers = 1;
 		dns_found(c, 0);
 		as = 0;
@@ -814,7 +817,7 @@ static void connected(void *c_)
 		    || !numeric_ipv6_address(h, NULL, NULL)
 #endif
 		    ) goto skip_numeric_address;
-		SSL_set_tlsext_host_name(c->ssl->ssl, h);
+		SSL_set_tlsext_host_name(c->ssl->ssl, cast_const_char h);
 skip_numeric_address:
 #endif
 		mem_free(h);
