@@ -16,6 +16,7 @@
 
 #undef read
 #undef write
+#undef pipe
 #undef close
 #undef select
 
@@ -42,18 +43,18 @@ static void sigbreak(int sig)
 static unsigned screen_x = 80;
 static unsigned screen_y = 25;
 
-int get_terminal_size(int fd, int *x, int *y)
+void get_terminal_size(int *x, int *y)
 {
 	*x = screen_x = ScreenCols();
 	*y = screen_y = ScreenRows();
-	return 0;
 }
 
-void handle_terminal_resize(int fd, void (*fn)(void))
+void handle_terminal_resize(void (*fn)(int, int), int *x, int *y)
 {
+	get_terminal_size(x, y);
 }
 
-void unhandle_terminal_resize(int fd)
+void unhandle_terminal_resize(void)
 {
 }
 
@@ -77,7 +78,7 @@ static unsigned char dos_mouse_buttons;
 static int dos_mouse_last_x;
 static int dos_mouse_last_y;
 static int dos_mouse_last_button;
-static ttime dos_mouse_time;
+static uttime dos_mouse_time;
 
 static struct links_event *dos_mouse_queue = DUMMY;
 static int dos_mouse_queue_n;
@@ -152,6 +153,7 @@ static void dos_mouse_enqueue(int x, int y, int b)
 		dos_mouse_queue_n--;
 		goto set_last;
 	}
+	if ((unsigned)dos_mouse_queue_n > (unsigned)MAXINT / sizeof(struct links_event) - 1) overalloc();
 	dos_mouse_queue = mem_realloc(dos_mouse_queue, (dos_mouse_queue_n + 1) * sizeof(struct links_event));
 set_last:
 	dos_mouse_queue[dos_mouse_queue_n].ev = EV_MOUSE;
@@ -178,7 +180,7 @@ static void dos_mouse_poll(void)
 	__dpmi_regs r;
 	dos_poll_break();
 	if (dos_mouse_initialized) {
-		if (dos_mouse_initialized == 1 && (uttime)get_time() - (uttime)dos_mouse_time > DOS_MOUSE_FLUSH_TIME)
+		if (dos_mouse_initialized == 1 && get_time() - dos_mouse_time > DOS_MOUSE_FLUSH_TIME)
 			dos_mouse_initialized = 2;
 		for (i = 0; i < dos_mouse_buttons; i++) {
 			if (dos_mouse_initialized == 1 && i > 0) {
@@ -245,7 +247,7 @@ void *handle_mouse(int cons, void (*fn)(void *, unsigned char *, int), void *dat
 	txt_mouse_handler = fn;
 	txt_mouse_data = data;
 
-	return (void *)1;
+	return DUMMY;
 }
 
 void unhandle_mouse(void *data)
@@ -367,6 +369,7 @@ process_ansi:
 				continue;
 			case 10:
 				cursor_y++;
+				/*-fallthrough*/
 			case 13:
 				cursor_x = 0;
 				break;
@@ -376,6 +379,7 @@ process_ansi:
 					case ')':
 						if (!size--) goto ret;
 						str++;
+						continue;
 					case '7':
 					default:
 						continue;
@@ -410,8 +414,8 @@ process_ansi:
 					case 'm':
 						while (buffer[0] >= '0' && buffer[0] < '9') {
 							unsigned long a;
-							unsigned char *e;
-							a = strtoul(cast_const_char buffer, (char **)(void *)&e, 10);
+							char *e;
+							a = strtoul(cast_const_char buffer, &e, 10);
 							if (*e == ';') e++;
 							memmove(buffer, e, strlen(cast_const_char e) + 1);
 							switch (a) {
@@ -514,19 +518,13 @@ static inline void pipe_wake(void)
 
 #include "vpipe.inc"
 
-int c_pipe(int *fd)
+int dos_pipe(int fd[2])
 {
 	int r = vpipe_create(fd);
-	/*printf("c_pipe: (%d) : %d,%d\n", r, fd[0], fd[1]);*/
+	/*printf("dos_pipe: (%d) : %d,%d\n", r, fd[0], fd[1]);*/
 	return r;
 }
 
-void set_nonblock(int fd)
-{
-	int rs;
-	EINTRLOOP(rs, fcntl(fd, F_SETFL, O_NONBLOCK));
-}
-
 int dos_read(int fd, void *buf, size_t size)
 {
 	int r;
@@ -777,6 +775,16 @@ void init_os(void)
 	int s, rs;
 	struct sigaction sa;
 
+#ifdef _STAT_INODE
+	_djstat_flags |= _STAT_INODE;
+#endif
+#ifdef _STAT_EXEC_EXT
+	_djstat_flags |= _STAT_EXEC_EXT;
+#endif
+#ifdef _STAT_EXEC_MAGIC
+	_djstat_flags |= _STAT_EXEC_MAGIC;
+#endif
+
 	init_seq_len = strlen(cast_const_char init_seq);
 
 	/* preload the packet driver */
