@@ -49,9 +49,9 @@ static int get_ftp_response(struct connection *c, struct read_buffer *rb, int pa
 	set_connection_timeout(c);
 	again:
 	for (l = 0; l < rb->len; l++) if (rb->data[l] == 10) {
-		unsigned char *e;
-		long k = strtoul(cast_const_char rb->data, (char **)(void *)&e, 10);
-		if (e != rb->data + 3 || k < 100 || k >= 1000) return -1;
+		char *e;
+		long k = strtoul(cast_const_char rb->data, &e, 10);
+		if (cast_uchar e != rb->data + 3 || k < 100 || k >= 1000) return -1;
 		if (*e == '-') {
 			int i;
 			for (i = 0; i < rb->len - 5; i++) {
@@ -576,17 +576,12 @@ static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsig
 	/*add_to_str(&str, &sl, cast_uchar "   ");*/
 
 	dir = 0;
-	if (!*d || *d == PARSE_MODE_VMS) for (ee = 0; ee + 1 < p && !WHITECHAR(buf[ee]); ee++) {
+	if (!*d || *d == PARSE_MODE_VMS) for (ee = 0; p - ee > 1 && !WHITECHAR(buf[ee]); ee++) {
 		if (buf[ee] == ';' && buf[ee + 1] >= '1' && buf[ee + 1] <= '9') {
 			if (!ee) goto skip_vms;
 			if (ee >= 4 && buf[ee - 4] == '.' && buf[ee - 3] == 'D' && buf[ee - 2] == 'I' && buf[ee - 1] == 'R') {
 				if (ee == 4) goto raw;
 				dir = 1, ee -= 4;
-			} else {
-				/*if (buf[ee + 1] == '1' && (ee + 2 == p || buf[ee + 2] < '0' || buf[ee + 2] > '9')) goto no_version;
-				ee += 2;
-				while (ee < p && buf[ee] >= '0' && buf[ee] < '9') ee++;
-				no_version:;*/
 			}
 			if (*d != PARSE_MODE_VMS && !root) {
 				add_to_str(&str, &sl, cast_uchar "<a href=\"../\">..</a>\n");
@@ -657,7 +652,7 @@ static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsig
 			    buf[pp - 2] >= '0' && buf[pp - 2] <= '5' &&
 			    buf[pp - 1] >= '0' && buf[pp - 1] <= '9') {
 				ppos = pp;
-				if (pp + 2 < p && buf[pp + 1] == ' ' && buf[pp + 2] != ' ')
+				if (p - pp > 2 && buf[pp + 1] == ' ' && buf[pp + 2] != ' ')
 					ppos++;
 			}
 		}
@@ -666,11 +661,11 @@ static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsig
 			goto done;
 		}
 
-		for (pp = 0; pp + 5 <= p; pp++)
+		for (pp = 0; p - pp >= 5; pp++)
 			if (!casecmp(&buf[pp], cast_uchar "<DIR>", 5)) {
 				pp += 4;
-				while (pp + 1 < p && WHITECHAR(buf[pp + 1])) pp++;
-				if (pp + 1 < p) goto done;
+				while (p - pp > 1 && WHITECHAR(buf[pp + 1])) pp++;
+				if (p - pp > 1) goto done;
 			}
 
 		bn = -1;
@@ -693,8 +688,8 @@ static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsig
 		}
 		if (bn >= 0) {
 			pp = bp + bn;
-			while (pp + 1 < p && WHITECHAR(buf[pp + 1])) pp++;
-			if (pp + 1 < p) goto done;
+			while (p - pp > 1 && WHITECHAR(buf[pp + 1])) pp++;
+			if (p - pp > 1) goto done;
 		}
 
 		for (pp = p - 1; pp >= 0; pp--) if (!WHITECHAR(buf[pp])) break;
