@@ -12,11 +12,16 @@
 
 #include "links.h"
 
-#ifdef HAVE_MATH_H
-#include <math.h>
-#endif
+struct gif_table_entry {
+	unsigned char end_char;
+	unsigned char garbage; /* This has nothing common to do with code table:
+			this is temporarily used for reverting strings :-) */
+	short pointer; /* points onto another entry in table, number 0...4095.
+			number -1 means it end there, the end_char is the last
+			number -2 means that this entry is no filled in yet. */
+};
 
-struct gif_decoder{
+struct gif_decoder {
 	unsigned char *color_map; /* NULL if no color map, otherwise a block of 768 bytes, red, green, blue,
 				     in sRGB, describing color slots 0...255. */
 	int state; /* State of the automatus finitus recognizing the GIF
@@ -253,7 +258,7 @@ accept_code(struct gif_decoder *deco,int c)
 {
  int k;
 
- if (c>4096||c<0) return; /* Erroneous code word will be ignored */
+ if (c >= 4096 || c < 0) return; /* Erroneous code word will be ignored */
  if (c==deco->CC) {
   init_table();
   return;
@@ -326,13 +331,11 @@ accept_byte(unsigned char c)
  * and the buffer must be formatted. */
 static void implant_transparent(struct gif_decoder *deco)
 {
-	if (deco->transparent>=0&&deco->transparent<(1<<deco->im_bpp)){
-		if (global_cimg->strip_optimized){
-			compute_background_8(deco->color_map+3*deco->transparent,
-				global_cimg);
-		}else{
-			memcpy(deco->color_map+3*deco->transparent
-				,deco->actual_line,3);
+	if (deco->transparent >= 0 && deco->transparent < (1 << deco->im_bpp)) {
+		if (global_cimg->strip_optimized) {
+			compute_background_8(global_cimg, deco->color_map + 3 * deco->transparent);
+		} else {
+			memcpy(deco->color_map + 3 * deco->transparent, deco->actual_line, 3);
 	        }
 	}
 }
@@ -442,7 +445,7 @@ gif_accept_byte(unsigned char c)
 			if (global_cimg->width && (unsigned)global_cimg->width * (unsigned)global_cimg->buffer_bytes_per_pixel / (unsigned)global_cimg->width != (unsigned)global_cimg->buffer_bytes_per_pixel) overalloc();
 			if ((unsigned)global_cimg->width * (unsigned)global_cimg->buffer_bytes_per_pixel > MAXINT) overalloc();
 			deco->actual_line=global_cimg->strip_optimized
-				?mem_alloc(global_cimg->width*global_cimg
+				?mem_alloc((size_t)global_cimg->width*global_cimg
 				->buffer_bytes_per_pixel)
 				:global_cimg->buffer;
 			if (deco->tbuf[8]&128){
