@@ -226,15 +226,16 @@ static unsigned char *set_mode_according_to_param(unsigned char *param)
 	} else {
 		unsigned long x, y;
 		int pl;
-		unsigned char *ptr = param, *end;
-		x = strtoul(cast_const_char ptr, (char **)(void *)&end, 10);
-		if (end == ptr || *end != 'x' || x >= MAXINT)
+		unsigned char *ptr = param;
+		char *end;
+		x = strtoul(cast_const_char ptr, &end, 10);
+		if (cast_uchar end == ptr || *end != 'x' || x >= MAXINT)
 			goto bad_param;
-		ptr = end + 1;
-		y = strtoul(cast_const_char ptr, (char **)(void *)&end, 10);
-		if (end == ptr || *end != 'x' || y >= MAXINT)
+		ptr = cast_uchar end + 1;
+		y = strtoul(cast_const_char ptr, &end, 10);
+		if (cast_uchar end == ptr || *end != 'x' || y >= MAXINT)
 			goto bad_param;
-		ptr = end + 1;
+		ptr = cast_uchar end + 1;
 		if (!strcmp(cast_const_char ptr, "16")) pl = 4;
 		else if (!strcmp(cast_const_char ptr, "256")) pl = 8;
 		else if (!casestrcmp(ptr, cast_uchar "32k") || !strcmp(cast_const_char ptr, "32768")) pl = 15;
@@ -566,6 +567,7 @@ struct graphics_driver grx_driver = {
 	grx_block,
 	grx_unblock,
 	grx_update_palette, /* set_palette */
+	NULL, /* get_real_colors */
 	NULL, /* set_title */
 	NULL, /* exec */
 	NULL, /* set_clipboard_text */
