@@ -107,7 +107,7 @@ static int ssl_set_private_paths(SSL_CTX *ctx)
 	unsigned char *path, *c;
 	int r;
 #ifdef OPENVMS
-	path = stracpy(g_argv[0]);
+	path = stracpy(cast_uchar g_argv[0]);
 #else
 	path = stracpy(path_to_exe);
 #endif
@@ -165,11 +165,11 @@ static void ssl_load_private_certificates(SSL_CTX *ctx)
 				unsigned char out[4];
 				out[0] = base64_chars[data[x] >> 2];
 				out[1] = base64_chars[((data[x] << 4) & 63) | (data[x + 1] >> 4)];
-				if (x + 1 < len)
+				if (len - x > 1)
 					out[2] = base64_chars[((data[x + 1] << 2) & 63) | (data[x + 2] >> 6)];
 				else
 					out[2] = '=';
-				if (x + 2 < len)
+				if (len - x > 2)
 					out[3] = base64_chars[data[x + 2] & 63];
 				else
 					out[3] = '=';
