@@ -78,7 +78,7 @@ retry:
 	EINTRLOOP(l, (int)write(itrm->sock_out, itrm->ev_queue, to_write));
 	if (l <= 0) {
 		if (to_write > 1) {
-			to_write >>= 2;
+			to_write >>= 1;
 			goto retry;
 		}
 		itrm_error(itrm);
@@ -161,11 +161,9 @@ static void send_term_sequence(int h, int flags)
 	done_draw();
 }
 
-static void resize_terminal(void)
+static void resize_terminal(int x, int y)
 {
 	struct links_event ev = { EV_RESIZE, 0, 0, 0 };
-	int x, y;
-	if (get_terminal_size(ditrm->std_out, &x, &y)) return;
 	ev.x = x;
 	ev.y = y;
 	queue_event(ditrm, (unsigned char *)&ev, sizeof(struct links_event));
@@ -354,15 +352,10 @@ static void setcooked(int ctl)
 
 void handle_trm(int std_in, int std_out, int sock_in, int sock_out, int ctl_in, void *init_string, int init_len)
 {
-	int x, y;
 	struct itrm *itrm;
-	struct links_event ev = { EV_INIT, 80, 24, 0 };
+	struct links_event ev = { EV_INIT, 0, 0, 0 };
 	unsigned char *ts;
 	int xwin, def_charset;
-	if (get_terminal_size(ctl_in, &x, &y)) {
-		error("ERROR: could not get terminal size");
-		return;
-	}
 	itrm = mem_alloc(sizeof(struct itrm));
 	itrm->queue_event = queue_event;
 	itrm->free_trm = free_trm;
@@ -380,9 +373,7 @@ void handle_trm(int std_in, int std_out, int sock_in, int sock_out, int ctl_in,
 	setraw(itrm->ctl_in, 1);
 	set_handlers(std_in, in_kbd, NULL, itrm);
 	if (sock_in != std_out) set_handlers(sock_in, in_sock, NULL, itrm);
-	ev.x = x;
-	ev.y = y;
-	handle_terminal_resize(ctl_in, resize_terminal);
+	handle_terminal_resize(resize_terminal, &ev.x, &ev.y);
 	queue_event(itrm, (unsigned char *)&ev, sizeof(struct links_event));
 	xwin = is_xterm() * ENV_XWIN + is_twterm() * ENV_TWIN + is_screen() * ENV_SCREEN + get_system_env();
 	itrm->flags = 0;
@@ -430,16 +421,17 @@ static void unblock_itrm_x(void *h)
 int unblock_itrm(int fd)
 {
 	struct itrm *itrm = ditrm;
+	int x, y;
 	if (!itrm) return -1;
 	if (setraw(itrm->ctl_in, 0)) return -1;
 	if (itrm->blocked != fd + 1) return -2;
 	itrm->blocked = 0;
 	send_init_sequence(itrm->std_out, itrm->flags);
 	set_handlers(itrm->std_in, in_kbd, NULL, itrm);
-	handle_terminal_resize(itrm->ctl_in, resize_terminal);
+	handle_terminal_resize(resize_terminal, &x, &y);
 	unblock_stdin();
 	itrm->mouse_h = handle_mouse(0, mouse_queue_event, itrm);
-	resize_terminal();
+	resize_terminal(x, y);
 	return 0;
 }
 
@@ -451,7 +443,7 @@ void block_itrm(int fd)
 	if (itrm->blocked) return;
 	itrm->blocked = fd + 1;
 	block_stdin();
-	unhandle_terminal_resize(itrm->ctl_in);
+	unhandle_terminal_resize();
 	if (itrm->mouse_h) unhandle_mouse(itrm->mouse_h), itrm->mouse_h = NULL;
 	send_term_sequence(itrm->std_out, itrm->flags);
 	setcooked(itrm->ctl_in);
@@ -463,7 +455,7 @@ static void free_trm(struct itrm *itrm)
 	if (!itrm) return;
 	set_window_title(itrm->orig_title);
 	if (itrm->orig_title) mem_free(itrm->orig_title), itrm->orig_title = NULL;
-	unhandle_terminal_resize(itrm->ctl_in);
+	unhandle_terminal_resize();
 	if (itrm->mouse_h) unhandle_mouse(itrm->mouse_h);
 	send_term_sequence(itrm->std_out, itrm->flags);
 	setcooked(itrm->ctl_in);
@@ -492,6 +484,16 @@ void fatal_tty_exit(void)
 #endif
 }
 
+static void refresh_terminal_size(void)
+{
+	if (!ditrm->blocked) {
+		int new_x, new_y;
+		unhandle_terminal_resize();
+		handle_terminal_resize(resize_terminal, &new_x, &new_y);
+		resize_terminal(new_x, new_y);
+	}
+}
+
 static void resize_terminal_x(unsigned char *text)
 {
 	int x, y;
@@ -501,7 +503,7 @@ static void resize_terminal_x(unsigned char *text)
 	x = atoi(cast_const_char text);
 	y = atoi(cast_const_char p);
 	resize_window(x, y);
-	resize_terminal();
+	refresh_terminal_size();
 }
 
 void dispatch_special(unsigned char *text)
@@ -636,8 +638,13 @@ static void kbd_timeout(void *itrm_)
 		internal("timeout on empty queue");
 		return;
 	}
+	if (itrm->kqueue[0] != 27) {
+		len = 1;
+		goto skip_esc;
+	}
 	itrm->queue_event(itrm, (unsigned char *)&ev, sizeof(struct links_event));
 	if (get_esc_code(itrm->kqueue, itrm->qlen, &code, &num, &len)) len = 1;
+	skip_esc:
 	itrm->qlen -= len;
 	memmove(itrm->kqueue, itrm->kqueue + len, itrm->qlen);
 	while (process_queue(itrm))
@@ -665,6 +672,7 @@ struct os2_key {
 };
 */
 
+#if defined(OS2) || defined(DOS)
 struct os2_key os2xtd[256] = {
 /* 0 */
 {0,0},
@@ -950,6 +958,7 @@ struct os2_key os2xtd[256] = {
 {0,0},
 /* 256 */
 };
+#endif
 
 static int xterm_button = -1;
 
@@ -1096,7 +1105,7 @@ static int process_queue(struct itrm *itrm)
 					case 201: itrm->flags &= ~BRACKETED_PASTE; break;
 					} break;
 				case 'R':
-					resize_terminal();
+					refresh_terminal_size();
 					break;
 				case 'M':
 				case '<':
@@ -1235,12 +1244,17 @@ do_alt:
 		}
 		goto l1;
 	} else if (itrm->kqueue[0] == 0) {
+#if defined(OS2) || defined(DOS)
 		if (itrm->qlen < 2) goto ret;
 		if (!(ev.x = os2xtd[itrm->kqueue[1]].x)) ev.x = -1;
 		ev.y = os2xtd[itrm->kqueue[1]].y;
 		el = 2;
 		/*printf("%02x - %02x %02x\n", (int)itrm->kqueue[1], ev.x, ev.y);*/
 		goto l1;
+#else
+		el = 1;
+		goto l1;
+#endif
 	}
 	el = 1;
 	ev.x = itrm->kqueue[0];
