@@ -6,7 +6,6 @@
 #include "links.h"
 
 struct translation {
-	int code;
 	const char *name;
 };
 
@@ -104,9 +103,18 @@ int get_default_language(void)
 	}
 
 	default_language = get_language_from_lang(cast_uchar "en");
+	if (default_language < 0)
+		internal("default language 'english' not found");
 	return default_language;
 }
 
+int get_current_language(void)
+{
+	if (current_language >= 0)
+		return current_language;
+	return get_default_language();
+}
+
 int get_default_charset(void)
 {
 	static int default_charset = -1;
@@ -119,7 +127,9 @@ int get_default_charset(void)
 	if (i >= 0)
 		goto ret_i;
 
-	lang = cast_uchar getenv("LANG");
+	lang = cast_uchar getenv("LC_CTYPE");
+	if (!lang)
+		lang = cast_uchar getenv("LANG");
 	if (!lang) {
 		i = 0;
 		goto ret_i;
@@ -130,7 +140,9 @@ int get_default_charset(void)
 		if (strlen(cast_const_char lang) > 5 && !casestrcmp(cast_uchar (strchr(cast_const_char lang, 0) - 5), cast_uchar "@euro")) {
 			p = cast_uchar "ISO-8859-15";
 		} else {
-			int def_lang = get_default_language();
+			int def_lang = get_language_from_lang(lang);
+			if (def_lang < 0)
+				def_lang = get_default_language();
 			p = cast_uchar translations[def_lang].t[T__DEFAULT_CHAR_SET].name;
 			if (!p)
 				p = cast_uchar "";
@@ -142,22 +154,26 @@ int get_default_charset(void)
 		i = 0;
 
 	ret_i:
+#ifndef ENABLE_UTF8
+	if (!F && i == utf8_table)
+		i = 0;
+#endif
 	default_charset = i;
 	return i;
 }
 
-int get_current_language(void)
+int get_commandline_charset(void)
 {
-	if (current_language >= 0)
-		return current_language;
-	return get_default_language();
+	return dump_codepage == -1 ? get_default_charset() : dump_codepage;
 }
 
-static inline int is_direct_text(unsigned char *text)
+static inline int get_text_index(unsigned char *text)
 {
-/* Do not compare to dummyarray directly - thwart some misoptimizations */
-	unsigned char * volatile dm = dummyarray;
-	return !(text >= dm && text < dm + T__N_TEXTS);
+	unsigned char * volatile dmv = dummyarray;
+	unsigned char * dm = dmv;
+	if (((my_uintptr_t)text >= (my_uintptr_t)dm && (my_uintptr_t)text < (my_uintptr_t)(dm + T__N_TEXTS)))
+		return (int)(text - dm);
+	return -1;
 }
 
 unsigned char *get_text_translation(unsigned char *text, struct terminal *term)
@@ -166,16 +182,18 @@ unsigned char *get_text_translation(unsigned char *text, struct terminal *term)
 	unsigned char *trn;
 	int charset;
 	int language_idx = get_current_language();
+	int idx;
 	if (!term) charset = 0;
 	else if (term->spec) charset = term_charset(term);
 	else charset = utf8_table;
-	if (is_direct_text(text)) return text;
+	idx = get_text_index(text);
+	if (idx < 0) return text;
 	if ((current_tra = translation_array[language_idx][charset])) {
 		unsigned char *tt;
-		if ((trn = current_tra[text - dummyarray])) return trn;
+		if ((trn = current_tra[idx])) return trn;
 		tr:
-		if (!(tt = cast_uchar translations[language_idx].t[text - dummyarray].name)) {
-			trn = cast_uchar translation_english[text - dummyarray].name;
+		if (!(tt = cast_uchar translations[language_idx].t[idx].name)) {
+			trn = cast_uchar translation_english[idx].name;
 		} else {
 			struct document_options l_opt;
 			memset(&l_opt, 0, sizeof(l_opt));
@@ -187,15 +205,15 @@ unsigned char *get_text_translation(unsigned char *text, struct terminal *term)
 				trn = tt;
 			}
 		}
-		current_tra[text - dummyarray] = trn;
+		current_tra[idx] = trn;
 	} else {
 		if (current_lang_charset && charset != current_lang_charset) {
-			current_tra = translation_array[language_idx][charset] = mem_alloc(sizeof (unsigned char **) * T__N_TEXTS);
-			memset(current_tra, 0, sizeof (unsigned char **) * T__N_TEXTS);
+			current_tra = translation_array[language_idx][charset] = mem_alloc(sizeof (unsigned char *) * T__N_TEXTS);
+			memset(current_tra, 0, sizeof (unsigned char *) * T__N_TEXTS);
 			goto tr;
 		}
-		if (!(trn = cast_uchar translations[language_idx].t[text - dummyarray].name)) {
-			trn = cast_uchar translation_english[text - dummyarray].name;
+		if (!(trn = cast_uchar translations[language_idx].t[idx].name)) {
+			trn = cast_uchar translation_english[idx].name;
 		}
 	}
 	return trn;
@@ -203,8 +221,9 @@ unsigned char *get_text_translation(unsigned char *text, struct terminal *term)
 
 unsigned char *get_english_translation(unsigned char *text)
 {
-	if (is_direct_text(text)) return text;
-	return cast_uchar translation_english[text - dummyarray].name;
+	int idx = get_text_index(text);
+	if (idx < 0) return text;
+	return cast_uchar translation_english[idx].name;
 }
 
 int n_languages(void)
@@ -224,10 +243,6 @@ void set_language(int l)
 	unsigned char *cp;
 	current_language = l;
 	l = get_current_language();
-	for (i = 0; i < T__N_TEXTS; i++) if (translations[l].t[i].code != i) {
-		internal("Bad table for language %s. Run script synclang.", translations[l].t[T__LANGUAGE].name);
-		return;
-	}
 	cp = cast_uchar translations[l].t[T__CHAR_SET].name;
 	i = get_cp_index(cp);
 	if (i == -1) {
