@@ -163,6 +163,15 @@
 
 #define ESC_TIMEOUT			200
 
+#ifdef OPENVMS
+#define TERMINAL_POLL_TIMEOUT		3000
+#else
+#define TERMINAL_POLL_TIMEOUT		1000
+#endif
+
+#define WIN32_REPAINT_IDLE_TIME		100
+#define WIN32_REPAINT_BUSY_TIME		1000
+
 #define DISPLAY_TIME_MIN		200
 #define DISPLAY_TIME_MAX_FIRST		1000
 #define DISPLAY_TIME			15
