@@ -18,8 +18,8 @@ struct smb_connection_info {
 	unsigned char text[1];
 };
 
-static void smb_got_data(struct connection *);
-static void smb_got_text(struct connection *);
+static void smb_got_data(void *);
+static void smb_got_text(void *);
 static void end_smb_connection(struct connection *);
 
 void smb_func(struct connection *c)
@@ -66,7 +66,7 @@ void smb_func(struct connection *c)
 	if ((p = cast_uchar strchr(cast_const_char data, '/'))) share = memacpy(data, p - data), dir = p + 1;
 	else if (*data) {
 		if (!c->cache) {
-			if (get_cache_entry(c->url, &c->cache)) {
+			if (get_connection_cache_entry(c)) {
 				mem_free(host);
 				mem_free(port);
 				mem_free(user);
@@ -80,7 +80,6 @@ void smb_func(struct connection *c)
 		}
 		if (c->cache->redirect) mem_free(c->cache->redirect);
 		c->cache->redirect = stracpy(c->url);
-		c->cache->redirect_get = 1;
 		add_to_strn(&c->cache->redirect, cast_uchar "/");
 		c->cache->incomplete = 0;
 		mem_free(host);
@@ -254,7 +253,7 @@ void smb_func(struct connection *c)
 			}
 			break;
 		default:
-			internal("unsuported smb client");
+			internal("unsupported smb client");
 		}
 		v[n++] = NULL;
 		EINTRLOOP(rs, execvp(cast_const_char v[0], (void *)v));
@@ -272,8 +271,8 @@ void smb_func(struct connection *c)
 	c->sock2 = pe[0];
 	EINTRLOOP(rs, close(po[1]));
 	EINTRLOOP(rs, close(pe[1]));
-	set_handlers(po[0], (void (*)(void *))smb_got_data, NULL, c);
-	set_handlers(pe[0], (void (*)(void *))smb_got_text, NULL, c);
+	set_handlers(po[0], smb_got_data, NULL, c);
+	set_handlers(pe[0], smb_got_text, NULL, c);
 	setcstate(c, S_CONN);
 }
 
@@ -323,7 +322,7 @@ static void smb_read_text(struct connection *c, int sock)
 {
 	int r;
 	struct smb_connection_info *si = c->info;
-	if ((unsigned)sizeof(struct smb_connection_info) + si->ntext + page_size + 2 > MAXINT) overalloc();
+	if ((size_t)sizeof(struct smb_connection_info) + si->ntext + page_size + 2 > (size_t)MAXINT) overalloc();
 	si = mem_realloc(si, sizeof(struct smb_connection_info) + si->ntext + page_size + 2);
 	c->info = si;
 	EINTRLOOP(r, (int)read(sock, si->text + si->ntext, page_size));
@@ -347,7 +346,7 @@ static void smb_read_text(struct connection *c, int sock)
 		int lasti = 0;
 		int i = 0;
 		si->text[si->ntext] = 0;
-		for (i = 0; i + 7 < si->ntext; i++) {
+		for (i = 0; si->ntext - i > 7; i++) {
 			nexti:
 			if ((si->text[i] == '\n' || si->text[i] == '\r') && (si->text[i + 1] == ' ' || (si->text[i + 1] >= '0' && si->text[i + 1] <= '9')) && ((si->text[i + 2] == ' ' && si->text[i + 1] == ' ') || (si->text[i + 2] >= '0' && si->text[i + 2] <= '9')) && (si->text[i + 3] >= '0' && si->text[i + 3] <= '9') && si->text[i + 4] == '%' && si->text[i + 5] == ' ' && si->text[i + 6] == '[') {
 				off_t position, total = 0; /* against warning */
@@ -379,8 +378,9 @@ static void smb_read_text(struct connection *c, int sock)
 	}
 }
 
-static void smb_got_data(struct connection *c)
+static void smb_got_data(void *c_)
 {
+	struct connection *c = (struct connection *)c_;
 	struct smb_connection_info *si = c->info;
 	unsigned char *buffer = mem_alloc(page_size);
 	int r;
@@ -409,7 +409,7 @@ static void smb_got_data(struct connection *c)
 	}
 	setcstate(c, S_TRANS);
 	if (!c->cache) {
-		if (get_cache_entry(c->url, &c->cache)) {
+		if (get_connection_cache_entry(c)) {
 			setcstate(c, S_OUT_OF_MEM);
 			abort_connection(c);
 			mem_free(buffer);
@@ -436,8 +436,9 @@ static void smb_got_data(struct connection *c)
 	mem_free(buffer);
 }
 
-static void smb_got_text(struct connection *c)
+static void smb_got_text(void *c_)
 {
+	struct connection *c = (struct connection *)c_;
 	smb_read_text(c, c->sock2);
 }
 
@@ -445,7 +446,7 @@ static void end_smb_connection(struct connection *c)
 {
 	struct smb_connection_info *si = c->info;
 	if (!c->cache) {
-		if (get_cache_entry(c->url, &c->cache)) {
+		if (get_connection_cache_entry(c)) {
 			setcstate(c, S_OUT_OF_MEM);
 			abort_connection(c);
 			return;
@@ -483,7 +484,6 @@ static void end_smb_connection(struct connection *c)
 		     strstr(cast_const_char si->text, "ERRbadfile") || sdir)) {
 			if (c->cache->redirect) mem_free(c->cache->redirect);
 			c->cache->redirect = stracpy(c->url);
-			c->cache->redirect_get = 1;
 			add_to_strn(&c->cache->redirect, cast_uchar "/");
 			c->cache->incomplete = 0;
 		} else {
@@ -598,7 +598,7 @@ static void end_smb_connection(struct connection *c)
 						int dir;
 						unsigned char *pp;
 						unsigned char *p = ls + 3;
-						while (p + 2 <= le2) {
+						while (le2 - p >= 2) {
 							if (p[0] == ' ' && p[1] == ' ') goto o;
 							p++;
 						}
