@@ -157,8 +157,7 @@ static void tiff_error_handler(const char* module, const char* fmt, va_list ap)
 
 static void flip_buffer(void *buf,int width,int height)
 {
-	if (htonl(0x12345678L)!=0x12345678L)	/* little endian --- ja to chci na intelu rychly!!! */
-	{
+	if (!big_endian) {	/* little endian --- ja to chci na intelu rychly!!! */
 #ifdef t4c
 		t4c* buffer=(t4c*)buf;
 		register t4c a,b;
@@ -184,9 +183,7 @@ static void flip_buffer(void *buf,int width,int height)
 			memcpy(tmp,p,w),memcpy(p,q,w),memcpy(q,tmp,w);
 		mem_free(tmp);
 #endif
-	}
-	else	/* big endian */
-	{
+	} else {	/* big endian */
 		unsigned char zakazany_uvolneni[4];
 		unsigned char* buffer=(unsigned char*)buf;
 		int w=width<<2; /* 4 bytes per pixel */
