@@ -123,6 +123,7 @@ void g_text_draw(struct f_data_c *fd, struct g_object *t_, int x, int y)
 		switch (form->type) {
 			struct style *inv;
 			int in, lid;
+			int sl, td;
 			case FC_RADIO:
 				if (link && fd->active && fd->vs->g_display_link && fd->vs->current_link == link - fd->f_data->links) inv = g_invert_style(t->style), in = 1;
 				else inv = t->style, in = 0;
@@ -147,15 +148,14 @@ void g_text_draw(struct f_data_c *fd, struct g_object *t_, int x, int y)
 			case FC_TEXT:
 			case FC_PASSWORD:
 			case FC_FILE:
-				/*
-				if (fs->state >= fs->vpos + form->size) fs->vpos = fs->state - form->size + 1;
-				if (fs->state < fs->vpos) fs->vpos = fs->state;
-				*/
 				if ((size_t)fs->vpos > strlen(cast_const_char fs->string)) fs->vpos = (int)strlen(cast_const_char fs->string);
-				while ((size_t)fs->vpos < strlen(cast_const_char fs->string) && textptr_diff(fs->string + fs->state, fs->string + fs->vpos, fd->f_data->opt.cp) >= form->size) {
+				sl = (int)strlen(cast_const_char fs->string);
+				td = textptr_diff(fs->string + fs->state, fs->string + fs->vpos, fd->f_data->opt.cp);
+				while (fs->vpos < sl && td >= form->size) {
 					unsigned char *p = fs->string + fs->vpos;
 					FWD_UTF_8(p);
 					fs->vpos = (int)(p - fs->string);
+					td--;
 				}
 				while (fs->vpos > fs->state) {
 					unsigned char *p = fs->string + fs->vpos;
@@ -1334,13 +1334,13 @@ void draw_title(struct f_data_c *f)
 	}
 	w = g_text_width(bfu_style_bw, title);
 	z = 0;
-	g_print_text(dev, 0, 0, bfu_style_bw, cast_uchar G_LEFT_ARROW, &z);
+	g_print_text(dev, 0, 0, !proxies.only_proxies ? bfu_style_bw : bfu_style_wb, cast_uchar G_LEFT_ARROW, &z);
 	f->ses->back_size = z;
 	b = (dev->size.x2 - w) - 16;
 	if (b < z) b = z;
-	drv->fill_area(dev, z, 0, b, G_BFU_FONT_SIZE, bfu_bg_color);
-	g_print_text(dev, b, 0, bfu_style_bw, title, &b);
-	drv->fill_area(dev, b, 0, dev->size.x2, G_BFU_FONT_SIZE, bfu_bg_color);
+	drv->fill_area(dev, z, 0, b, G_BFU_FONT_SIZE, !proxies.only_proxies ? bfu_bg_color : bfu_fg_color);
+	g_print_text(dev, b, 0, !proxies.only_proxies ? bfu_style_bw : bfu_style_wb, title, &b);
+	drv->fill_area(dev, b, 0, dev->size.x2, G_BFU_FONT_SIZE, !proxies.only_proxies ? bfu_bg_color : bfu_fg_color);
 	mem_free(title);
 }
 
